from datetime import timedelta, datetime
from six import iteritems, itervalues

from .search import Search
from .aggs import A
from .utils import AttrDict
from .response import Response
from .query import Terms, Nested, Range, MatchAll

__all__ = [
    'FacetedSearch', 'HistogramFacet', 'TermsFacet', 'DateHistogramFacet', 'RangeFacet',
    'NestedFacet',
]

class Facet(object):
    """
    A facet on faceted search. Wraps and aggregation and provides functionality
    to create a filter for selected values and return a list of facet values
    from the result of the aggregation.
    """
    agg_type = None

    def __init__(self, metric=None, metric_sort='desc', **kwargs):
        self.filter_values = ()
        self._params = kwargs
        self._metric = metric
        if metric and metric_sort:
            self._params['order'] = {'metric': metric_sort}

    def get_aggregation(self):
        """
        Return the aggregation object.
        """
        agg = A(self.agg_type, **self._params)
        if self._metric:
            agg.metric('metric', self._metric)
        return agg

    def add_filter(self, filter_values):
        """
        Construct a filter.
        """
        if not filter_values:
            return

        f = self.get_value_filter(filter_values[0])
        for v in filter_values[1:]:
            f |= self.get_value_filter(v)
        return f

    def get_value_filter(self, filter_value):
        """
        Construct a filter for an individual value
        """
        pass

    def is_filtered(self, key, filter_values):
        """
        Is a filter active on the given key.
        """
        return key in filter_values

    def get_value(self, bucket):
        """
        return a value representing a bucket. Its key as default.
        """
        return bucket['key']

    def get_metric(self, bucket):
        """
        Return a metric, by default doc_count for a bucket.
        """
        if self._metric:
            return bucket['metric']['value']
        return bucket['doc_count']

    def get_values(self, data, filter_values):
        """
        Turn the raw bucket data into a list of tuples containing the key,
        number of documents and a flag indicating whether this value has been
        selected or not.
        """
        out = []
        for bucket in data.buckets:
            key = self.get_value(bucket)
            out.append((
                key,
                self.get_metric(bucket),
                self.is_filtered(key, filter_values)
            ))
        return out


class TermsFacet(Facet):
    agg_type = 'terms'

    def add_filter(self, filter_values):
        """ Create a terms filter instead of bool containing term filters.  """
        if filter_values:
            return Terms(**{self._params['field']: filter_values})


class RangeFacet(Facet):
    agg_type = 'range'

    def _range_to_dict(self, range):
        key, range = range
        out = {'key': key}
        if range[0] is not None:
            out['from'] = range[0]
        if range[1] is not None:
            out['to'] = range[1]
        return out

    def __init__(self, ranges, **kwargs):
        super(RangeFacet, self).__init__(**kwargs)
        self._params['ranges'] = list(map(self._range_to_dict, ranges))
        self._params['keyed'] = False
        self._ranges = dict(ranges)

    def get_value_filter(self, filter_value):
        f, t = self._ranges[filter_value]
        limits = {}
        if f is not None:
            limits['gte'] = f
        if t is not None:
            limits['lt'] = t

        return Range(**{
            self._params['field']: limits
        })

class HistogramFacet(Facet):
    agg_type = 'histogram'

    def get_value_filter(self, filter_value):
        return Range(**{
            self._params['field']: {
                'gte': filter_value,
                'lt': filter_value + self._params['interval']
            }
        })


class DateHistogramFacet(Facet):
    agg_type = 'date_histogram'

    DATE_INTERVALS = {
        'month': lambda d: (d+timedelta(days=32)).replace(day=1),
        'week': lambda d: d+timedelta(days=7),
        'day': lambda d: d+timedelta(days=1),
        'hour': lambda d: d+timedelta(hours=1),
    }

    def __init__(self, **kwargs):
        kwargs.setdefault("min_doc_count", 0)
        super(DateHistogramFacet, self).__init__(**kwargs)

    def get_value(self, bucket):
        if not isinstance(bucket['key'], datetime):
            # Elasticsearch returns key=None instead of 0 for date 1970-01-01,
            # so we need to set key to 0 to avoid TypeError exception
            if bucket['key'] is None:
                bucket['key'] = 0
            # Preserve milliseconds in the datetime
            return datetime.utcfromtimestamp(int(bucket['key']) / 1000.0)
        else:
            return bucket['key']

    def get_value_filter(self, filter_value):
        return Range(**{
            self._params['field']: {
                'gte': filter_value,
                'lt': self.DATE_INTERVALS[self._params['interval']](filter_value)
            }
        })

class NestedFacet(Facet):
    agg_type = 'nested'

    def __init__(self, path, nested_facet):
        self._path = path
        self._inner = nested_facet
        super(NestedFacet, self).__init__(path=path, aggs={'inner': nested_facet.get_aggregation()})

    def get_values(self, data, filter_values):
        return self._inner.get_values(data.inner, filter_values)

    def add_filter(self, filter_values):
        inner_q = self._inner.add_filter(filter_values)
        if inner_q:
            return Nested(path=self._path, query=inner_q)

class FacetedResponse(Response):
    @property
    def query_string(self):
        return self._faceted_search._query

    @property
    def facets(self):
        if not hasattr(self, '_facets'):
            super(AttrDict, self).__setattr__('_facets', AttrDict({}))
            for name, facet in iteritems(self._faceted_search.facets):
                self._facets[name] = facet.get_values(
                    getattr(getattr(self.aggregations, '_filter_' + name), name),
                    self._faceted_search.filter_values.get(name, ())
                )
        return self._facets


class FacetedSearch(object):
    """
    Abstraction for creating faceted navigation searches that takes care of
    composing the queries, aggregations and filters as needed as well as
    presenting the results in an easy-to-consume fashion::

        class BlogSearch(FacetedSearch):
            index = 'blogs'
            doc_types = [Blog, Post]
            fields = ['title^5', 'category', 'description', 'body']

            facets = {
                'type': TermsFacet(field='_type'),
                'category': TermsFacet(field='category'),
                'weekly_posts': DateHistogramFacet(field='published_from', interval='week')
            }

            def search(self):
                ' Override search to add your own filters '
                s = super(BlogSearch, self).search()
                return s.filter('term', published=True)

        # when using:
        blog_search = BlogSearch("web framework", filters={"category": "python"})

        # supports pagination
        blog_search[10:20]

        response = blog_search.execute()

        # easy access to aggregation results:
        for category, hit_count, is_selected in response.facets.category:
            print(
                "Category %s has %d hits%s." % (
                    category,
                    hit_count,
                    ' and is chosen' if is_selected else ''
                )
            )

    """
    index = None
    doc_types = None
    fields = ('*', )
    facets = {}
    using = 'default'

    def __init__(self, query=None, filters={}, sort=()):
        """
        :arg query: the text to search for
        :arg filters: facet values to filter
        :arg sort: sort information to be passed to :class:`~elasticsearch_dsl.Search`
        """
        self._query = query
        self._filters = {}
        self._sort = sort
        self.filter_values = {}
        for name, value in iteritems(filters):
            self.add_filter(name, value)

        self._s = self.build_search()

    def count(self):
        return self._s.count()

    def __getitem__(self, k):
        self._s = self._s[k]
        return self

    def __iter__(self):
        return iter(self._s)

    def add_filter(self, name, filter_values):
        """
        Add a filter for a facet.
        """
        # normalize the value into a list
        if not isinstance(filter_values, (tuple, list)):
            if filter_values is None:
                return
            filter_values = [filter_values, ]

        # remember the filter values for use in FacetedResponse
        self.filter_values[name] = filter_values

        # get the filter from the facet
        f = self.facets[name].add_filter(filter_values)
        if f is None:
            return

        self._filters[name] = f

    def search(self):
        """
        Returns the base Search object to which the facets are added.

        You can customize the query by overriding this method and returning a
        modified search object.
        """
        s = Search(doc_type=self.doc_types, index=self.index, using=self.using)
        return s.response_class(FacetedResponse)

    def query(self, search, query):
        """
        Add query part to ``search``.

        Override this if you wish to customize the query used.
        """
        if query:
            return search.query('multi_match', fields=self.fields, query=query)
        return search

    def aggregate(self, search):
        """
        Add aggregations representing the facets selected, including potential
        filters.
        """
        for f, facet in iteritems(self.facets):
            agg = facet.get_aggregation()
            agg_filter = MatchAll()
            for field, filter in iteritems(self._filters):
                if f == field:
                    continue
                agg_filter &= filter
            search.aggs.bucket(
                '_filter_' + f,
                'filter',
                filter=agg_filter
            ).bucket(f, agg)

    def filter(self, search):
        """
        Add a ``post_filter`` to the search request narrowing the results based
        on the facet filters.
        """
        if not self._filters:
            return search

        post_filter = MatchAll()
        for f in itervalues(self._filters):
            post_filter &= f
        return search.post_filter(post_filter)

    def highlight(self, search):
        """
        Add highlighting for all the fields
        """
        return search.highlight(*(f if '^' not in f else f.split('^', 1)[0]
                                  for f in self.fields))

    def sort(self, search):
        """
        Add sorting information to the request.
        """
        if self._sort:
            search = search.sort(*self._sort)
        return search

    def build_search(self):
        """
        Construct the ``Search`` object.
        """
        s = self.search()
        s = self.query(s, self._query)
        s = self.filter(s)
        s = self.highlight(s)
        s = self.sort(s)
        self.aggregate(s)
        return s

    def execute(self):
        """
        Execute the search and return the response.
        """
        r = self._s.execute()
        r._faceted_search = self
        return r



try:
    import collections.abc as collections_abc  # only works on python 3.3+
except ImportError:
    import collections as collections_abc

from .utils import DslBase
from .response.aggs import BucketData, FieldBucketData, AggResponse, TopHitsData

def A(name_or_agg, filter=None, **params):
    if filter is not None:
        if name_or_agg != 'filter':
            raise ValueError("Aggregation %r doesn't accept positional argument 'filter'." % name_or_agg)
        params['filter'] = filter

    # {"terms": {"field": "tags"}, "aggs": {...}}
    if isinstance(name_or_agg, collections_abc.Mapping):
        if params:
            raise ValueError('A() cannot accept parameters when passing in a dict.')
        # copy to avoid modifying in-place
        agg = name_or_agg.copy()
        # pop out nested aggs
        aggs = agg.pop('aggs', None)
        # pop out meta data
        meta = agg.pop('meta', None)
        # should be {"terms": {"field": "tags"}}
        if len(agg) != 1:
            raise ValueError('A() can only accept dict with an aggregation ({"terms": {...}}). '
                             'Instead it got (%r)' % name_or_agg)
        agg_type, params = agg.popitem()
        if aggs:
            params = params.copy()
            params['aggs'] = aggs
        if meta:
            params = params.copy()
            params['meta'] = meta
        return Agg.get_dsl_class(agg_type)(_expand__to_dot=False, **params)

    # Terms(...) just return the nested agg
    elif isinstance(name_or_agg, Agg):
        if params:
            raise ValueError('A() cannot accept parameters when passing in an Agg object.')
        return name_or_agg

    # "terms", field="tags"
    return Agg.get_dsl_class(name_or_agg)(**params)

class Agg(DslBase):
    _type_name = 'agg'
    _type_shortcut = staticmethod(A)
    name = None

    def __contains__(self, key):
        return False

    def to_dict(self):
        d = super(Agg, self).to_dict()
        if 'meta' in d[self.name]:
            d['meta'] = d[self.name].pop('meta')
        return d

    def result(self, search, data):
        return AggResponse(self, search, data)


class AggBase(object):
    _param_defs = {
        'aggs': {'type': 'agg', 'hash': True},
    }
    def __contains__(self, key):
        return key in self._params.get('aggs', {})

    def __getitem__(self, agg_name):
        agg = self._params.setdefault('aggs', {})[agg_name]  # propagate KeyError

        # make sure we're not mutating a shared state - whenever accessing a
        # bucket, return a shallow copy of it to be safe
        if isinstance(agg, Bucket):
            agg = A(agg.name, **agg._params)
            # be sure to store the copy so any modifications to it will affect us
            self._params['aggs'][agg_name] = agg

        return agg

    def __setitem__(self, agg_name, agg):
        self.aggs[agg_name] = A(agg)

    def __iter__(self):
        return iter(self.aggs)

    def _agg(self, bucket, name, agg_type, *args, **params):
        agg = self[name] = A(agg_type, *args, **params)

        # For chaining - when creating new buckets return them...
        if bucket:
            return agg
        # otherwise return self._base so we can keep chaining
        else:
            return self._base

    def metric(self, name, agg_type, *args, **params):
        return self._agg(False, name, agg_type, *args, **params)

    def bucket(self, name, agg_type, *args, **params):
        return self._agg(True, name, agg_type, *args, **params)

    def pipeline(self, name, agg_type, *args, **params):
        return self._agg(False, name, agg_type, *args, **params)

    def result(self, search, data):
        return BucketData(self, search, data)


class Bucket(AggBase, Agg):
    def __init__(self, **params):
        super(Bucket, self).__init__(**params)
        # remember self for chaining
        self._base = self

    def to_dict(self):
        d = super(AggBase, self).to_dict()
        if 'aggs' in d[self.name]:
            d['aggs'] = d[self.name].pop('aggs')
        return d

class Filter(Bucket):
    name = 'filter'
    _param_defs = {
        'filter': {'type': 'query'},
        'aggs': {'type': 'agg', 'hash': True},
    }

    def __init__(self, filter=None, **params):
        if filter is not None:
            params['filter'] = filter
        super(Filter, self).__init__(**params)

    def to_dict(self):
        d = super(Filter, self).to_dict()
        d[self.name].update(d[self.name].pop('filter', {}))
        return d

class Pipeline(Agg):
    pass

# bucket aggregations
class Filters(Bucket):
    name = 'filters'
    _param_defs = {
        'filters': {'type': 'query', 'hash': True},
        'aggs': {'type': 'agg', 'hash': True},
    }

class Children(Bucket):
    name = 'children'

class DateHistogram(Bucket):
    name = 'date_histogram'

    def result(self, search, data):
        return FieldBucketData(self, search, data)

class DateRange(Bucket):
    name = 'date_range'

class GeoDistance(Bucket):
    name = 'geo_distance'

class GeohashGrid(Bucket):
    name = 'geohash_grid'

class GeoCentroid(Bucket):
    name = 'geo_centroid'

class Global(Bucket):
    name = 'global'

class Histogram(Bucket):
    name = 'histogram'
    def result(self, search, data):
        return FieldBucketData(self, search, data)

class IPRange(Bucket):
    name = 'ip_range'

class Missing(Bucket):
    name = 'missing'

class Nested(Bucket):
    name = 'nested'

class Range(Bucket):
    name = 'range'

class ReverseNested(Bucket):
    name = 'reverse_nested'

class SignificantTerms(Bucket):
    name = 'significant_terms'

class SignificantText(Bucket):
    name = 'significant_text'

class Terms(Bucket):
    name = 'terms'

    def result(self, search, data):
        return FieldBucketData(self, search, data)

class Sampler(Bucket):
    name = 'sampler'

class DiversifiedSampler(Bucket):
    name = 'diversified_sampler'

class Composite(Bucket):
    name = 'composite'
    _param_defs = {
        'sources': {'type': 'agg', 'hash': True, 'multi': True},
        'aggs': {'type': 'agg', 'hash': True},
    }

# metric aggregations
class TopHits(Agg):
    name = 'top_hits'

    def result(self, search, data):
        return TopHitsData(self, search, data)

class Avg(Agg):
    name = 'avg'

class Cardinality(Agg):
    name = 'cardinality'

class ExtendedStats(Agg):
    name = 'extended_stats'

class GeoBounds(Agg):
    name = 'geo_bounds'

class Max(Agg):
    name = 'max'

class Min(Agg):
    name = 'min'

class Percentiles(Agg):
    name = 'percentiles'

class PercentileRanks(Agg):
    name = 'percentile_ranks'

class ScriptedMetric(Agg):
    name = 'scripted_metric'

class Stats(Agg):
    name = 'stats'

class Sum(Agg):
    name = 'sum'

class ValueCount(Agg):
    name = 'value_count'

# pipeline aggregations
class AvgBucket(Pipeline):
    name = 'avg_bucket'

class BucketScript(Pipeline):
    name = 'bucket_script'

class BucketSelector(Pipeline):
    name = 'bucket_selector'

class CumulativeSum(Pipeline):
    name = 'cumulative_sum'

class Derivative(Pipeline):
    name = 'derivative'

class ExtendedStatsBucket(Pipeline):
    name = 'extended_stats_bucket'

class MaxBucket(Pipeline):
    name = 'max_bucket'

class MinBucket(Pipeline):
    name = 'min_bucket'

class MovingAvg(Pipeline):
    name = 'moving_avg'

class PercentilesBucket(Pipeline):
    name = 'percentiles_bucket'

class SerialDiff(Pipeline):
    name = 'serial_diff'

class StatsBucket(Pipeline):
    name = 'stats_bucket'

class SumBucket(Pipeline):
    name = 'sum_bucket'

class BucketSort(Pipeline):
    name = 'bucket_sort'



from ..utils import AttrDict, AttrList
from . import Response, AggResponse

class Bucket(AggResponse):
    def __init__(self, aggs, search, data, field=None):
        super(Bucket, self).__init__(aggs, search, data)

class FieldBucket(Bucket):
    def __init__(self, aggs, search, data, field=None):
        if field:
            data['key'] = field.deserialize(data['key'])
        super(FieldBucket, self).__init__(aggs, search, data, field)

class BucketData(AggResponse):
    _bucket_class = Bucket
    def _wrap_bucket(self, data):
        return self._bucket_class(self._meta['aggs'], self._meta['search'],
                                  data, field=self._meta.get('field'))

    def __iter__(self):
        return iter(self.buckets)

    def __len__(self):
        return len(self.buckets)

    def __getitem__(self, key):
        if isinstance(key, (int, slice)):
            return self.buckets[key]
        return super(BucketData, self).__getitem__(key)

    @property
    def buckets(self):
        if not hasattr(self, '_buckets'):
            field = getattr(self._meta['aggs'], 'field', None)
            if field:
                self._meta['field'] = self._meta['search']._resolve_field(field)
            bs = self._d_['buckets']
            if isinstance(bs, list):
                bs = AttrList(bs, obj_wrapper=self._wrap_bucket)
            else:
                bs = AttrDict(dict((k, self._wrap_bucket(bs[k])) for k in bs))
            super(AttrDict, self).__setattr__('_buckets', bs)
        return self._buckets

class FieldBucketData(BucketData):
    _bucket_class = FieldBucket

class TopHitsData(Response):
    def __init__(self, agg, search, data):
        super(AttrDict, self).__setattr__('meta', AttrDict({'agg': agg, 'search': search}))
        super(TopHitsData, self).__init__(search, data)



try:
    import collections.abc as collections_abc  # only works on python 3.3+
except ImportError:
    import collections as collections_abc

from .utils import DslBase

def SF(name_or_sf, **params):
    # {"script_score": {"script": "_score"}, "filter": {}}
    if isinstance(name_or_sf, collections_abc.Mapping):
        if params:
            raise ValueError('SF() cannot accept parameters when passing in a dict.')
        kwargs = {}
        sf = name_or_sf.copy()
        for k in ScoreFunction._param_defs:
            if k in name_or_sf:
                kwargs[k] = sf.pop(k)

        # not sf, so just filter+weight, which used to be boost factor
        if not sf:
            name = 'boost_factor'
        # {'FUNCTION': {...}}
        elif len(sf) == 1:
            name, params = sf.popitem()
        else:
            raise ValueError('SF() got an unexpected fields in the dictionary: %r' % sf)

        # boost factor special case, see elasticsearch #6343
        if not isinstance(params, collections_abc.Mapping):
            params = {'value': params}

        # mix known params (from _param_defs) and from inside the function
        kwargs.update(params)
        return ScoreFunction.get_dsl_class(name)(**kwargs)

    # ScriptScore(script="_score", filter=Q())
    if isinstance(name_or_sf, ScoreFunction):
        if params:
            raise ValueError('SF() cannot accept parameters when passing in a ScoreFunction object.')
        return name_or_sf

    # "script_score", script="_score", filter=Q()
    return ScoreFunction.get_dsl_class(name_or_sf)(**params)

class ScoreFunction(DslBase):
    _type_name = 'score_function'
    _type_shortcut = staticmethod(SF)
    _param_defs = {
        'query': {'type': 'query'},
        'filter': {'type': 'query'},
        'weight': {}
    }
    name = None

    def to_dict(self):
        d = super(ScoreFunction, self).to_dict()
        # filter and query dicts should be at the same level as us
        for k in self._param_defs:
            if k in d[self.name]:
                d[k] = d[self.name].pop(k)
        return d

class ScriptScore(ScoreFunction):
    name = 'script_score'

class BoostFactor(ScoreFunction):
    name = 'boost_factor'

    def to_dict(self):
        d = super(BoostFactor, self).to_dict()
        if 'value' in d[self.name]:
            d[self.name] = d[self.name].pop('value')
        else:
            del d[self.name]
        return d

class RandomScore(ScoreFunction):
    name = 'random_score'

class FieldValueFactor(ScoreFunction):
    name = 'field_value_factor'

class Linear(ScoreFunction):
    name = 'linear'

class Gauss(ScoreFunction):
    name = 'gauss'

class Exp(ScoreFunction):
    name = 'exp'

